#include "spheroidal.h"

// -------------------------------------------------------------------------------
void NonLinPotentials(parameters par, int idom, int i, int j, coordinate xx, coordinate yy, 
	derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *Laplace, double *grad_grad)
{	// Calculates the potential values, first and second (x,y)-derivatives as well as
	// the Laplacian of U at a given spatial grid point; for domain 4 see pages 12, 13
	double x = xx.d0, y = yy.d0;
	int ipot, m,n;
		
	v_To_V(par, idom, i, j, v, w, V);
	pot_st_To_xy(par, xx, yy, V);

	for(ipot=0; ipot<NPOT; ipot ++) 
		Laplace[ipot] = 4.*(x*V.d11[ipot]+y*V.d22[ipot]+V.d1[ipot]+0.5*V.d2[ipot]);

	for (m=0; m<NPOT; m++){
		for (n=m; n<NPOT; n++){
			grad_grad[GradIndex(m,n)] = 4.*(x*V.d1[m]*V.d1[n]+y*V.d2[m]*V.d2[n]);
		}	
	}
}
// -------------------------------------------------------------------------------
void LinPotentials(parameters par, int idom, int i, int j, coordinate  xx,  coordinate yy, 
	derivs_2D v,  derivs_1D w,  derivs_1D G,  derivs_2D V, double *Laplace,  double *grad_grad,
	coordinate Dxx, coordinate Dyy, derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV,
	double *DLaplace, double *Dgrad_grad)
{	// Linear version of 'NonLinPotentials'
	double x = xx.d0, y = yy.d0, Dx = Dxx.d0, Dy = Dyy.d0;
	int ipot, m, n;
		
	v_To_V(par, idom, i, j,  v,  w,  V);
	v_To_V(par, idom, i, j, Dv, Dw, DV);
	D_pot_st_To_xy(par, xx, yy, V, Dxx, Dyy, DV);
	for(ipot=0; ipot<NPOT; ipot ++){
		Laplace[ipot]   = 4.*(x* V.d11[ipot]               +y* V.d22[ipot] 
								+  V.d1[ipot] + 0.5*V.d2[ipot]);
		DLaplace[ipot]  = 4.*(x*DV.d11[ipot]+Dx*V.d11[ipot]+y*DV.d22[ipot]+Dy*V.d22[ipot]
								+ DV.d1[ipot] + 0.5*DV.d2[ipot]);
	}
	for (m=0; m<NPOT; m++){
		for (n=m; n<NPOT; n++){
			grad_grad[GradIndex(m,n)]  = 4.*(x*V.d1[m]*V.d1[n]+y*V.d2[m]*V.d2[n]);
			Dgrad_grad[GradIndex(m,n)] = 4.*(Dx*V.d1[m]*V.d1[n]+x*DV.d1[m]*V.d1[n]+x*V.d1[m]*DV.d1[n]
			                               + Dy*V.d2[m]*V.d2[n]+y*DV.d2[m]*V.d2[n]+y*V.d2[m]*DV.d2[n]);
		}	
	}
}
// -------------------------------------------------------------------------------
void NonLinFieldEqns(parameters par, int idom, int i, int j,
		derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F)
{	// Nonlinear field equations and regularity of V_yy at s*(1-s)=0 in domain 4
	
	int nt=par.nt[idom], Nt=nt-1, ipot;
	
	if (idom == NDOM-1 && j == Nt){  // V(s, t=1) = 0 for 0 <= s < 1
		for(ipot=0; ipot<NPOT; ipot ++){
			int indx = Index(par,idom,ipot,i,j);
			F[indx]  = v.d0[indx]; 
		}
	}
	else{// Nonlinear field equations
		double Laplace[NPOT], grad_grad[NPOT*NPOT], 
			x, y, u, ux, uy, B, Bx, By, om, omx, omy, mu, mux, muy, aux0, aux1;
		coordinate xx, yy;

		coord_st_To_xy(par, idom, i, j, G, &xx, &yy);
		NonLinPotentials(par, idom, i, j, xx, yy, v, w, G, V, Laplace, grad_grad);
		x = xx.d0;    y = yy.d0;
		u = V.d0[0]; ux = V.d1[0]; uy = V.d2[0];
		B = V.d0[1]; Bx = V.d1[1]; By = V.d2[1];
		om= V.d0[2]; omx= V.d1[2]; omy= V.d2[2];
		mu= V.d0[3]; mux= V.d1[3]; muy= V.d2[3];

		// The equations [Bardeen, 1971, <Relativistic Disks>, p. 363]: 
		//    \nabla*(\nabla U)    = U_rho_rho   + U_zeta_zeta + U_rho/rho 
		//    (\nabla U)*(nabla V) = U_rho*V_rho + U_zeta*V_zeta
					
		//    (1) Equation (II.12)*e^(2*mu) with nu = u + ln(B)
		//    (2) Equation (II.13)*e^(2*mu)/W
		//    (3) Equation (II.14)*e^(2*mu)

		aux0 = -0.5*x*exp(-4.*u)*grad_grad[GradIndex(2,2)]/B;
		aux1 = -2.*Bx + grad_grad[GradIndex(0,1)] + aux0;
		F[Index(par,idom,0,i,j)] = Laplace[0] + aux1/B;

		F[Index(par,idom,1,i,j)] = Laplace[1] + 2.*Bx;
		
		F[Index(par,idom,2,i,j)] = Laplace[2] + 4.*omx - grad_grad[GradIndex(1,2)]/B
			- 4.*grad_grad[GradIndex(0,2)];
		
		F[Index(par,idom,3,i,j)] = Laplace[3] - 2.*mux - 2.*ux - 2.*Bx/B
				+ grad_grad[GradIndex(0,0)] + grad_grad[GradIndex(0,1)]/B
				+ 0.5*aux0/B;
											
		if (idom == NDOM-1){ // Matter terms if in ring
			double A1,  A2,  diff,  diff2,  e2V,  e2mu,  hat_vel,  vel2,  h,  p,  eps,  muB;

			A1    = x*exp(-2.*u); 
			A2    = B*B*exp(2.*u);
			diff  = par.Omega - om;
			diff2 = diff*diff;
			e2V   = A2-A1*diff2;
			e2mu  = exp(2.*mu);

			hat_vel  = exp(-2.*u)*diff/B;
			vel2     = x*hat_vel*hat_vel;
			h        = par.V0 - 0.5*log(e2V);      // h    = V0 - V
			eos_quantities(par, h, &p, &eps, &muB); // h(p) = Integrate[1/(eps(pp)+pp), {pp, 0, p}]

			aux0        = (eps+p)*(1.+vel2)/(1.-vel2) - 2.*p;
			aux1        = (eps+p)*hat_vel/(1.-vel2);

			F[Index(par,idom,0,i,j)] -=   4.*Pi*  e2mu*aux0;
			F[Index(par,idom,1,i,j)] -=  16.*Pi*B*e2mu*p;
			F[Index(par,idom,2,i,j)] -= -16.*Pi*B*e2mu*exp(2.*u)*aux1;
			F[Index(par,idom,3,i,j)] -=  -4.*Pi*  e2mu*(eps+p);
		}
	}
}
// -------------------------------------------------------------------------------
void LinFieldEqns(parameters par, parameters Dpar, int idom, int i, int j,
	derivs_2D  v, derivs_1D  w, derivs_1D  G, derivs_2D  V,
	derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX)
{	// Linear version of 'NonLinPotentials'
	int nt=par.nt[idom], Nt=nt-1, ipot;
	
	if (idom == NDOM-1 && j == Nt){  // V(s, t=1) = 0 for 0 <= s < 1
		for(ipot=0; ipot<NPOT; ipot ++){
			int indx   = Index(par,idom,ipot,i,j);
			JDX[indx]  = Dv.d0[indx];
		}
	}
	else{
		double Laplace[NPOT], grad_grad[NPOT*NPOT], DLaplace[NPOT], Dgrad_grad[NPOT*NPOT],
			 x,  y,  u,  ux,  uy,  B,  Bx,  By,  om,  omx,  omy,  mu,  mux,  muy,  aux0,  aux1,
			Dx, Dy, Du, Dux, Duy, DB, DBx, DBy, Dom, Domx, Domy, Dmu, Dmux, Dmuy, Daux0, Daux1;
		coordinate xx, yy, Dxx, Dyy;

		Dcoord_st_To_xy(par, Dpar, idom, i, j, G, DG, &xx, &yy, &Dxx, &Dyy);
		
		LinPotentials(par, idom, i, j, xx,  yy,  v,  w,  G,  V,  Laplace,   grad_grad,
                                      Dxx, Dyy, Dv, Dw, DG, DV, DLaplace,  Dgrad_grad);
		x  = xx.d0;    y  = yy.d0;    Dx = Dxx.d0;   Dy = Dyy.d0;
		u  = V.d0[0];  ux = V.d1[0];  uy = V.d2[0];  Du = DV.d0[0];  Dux = DV.d1[0];  Duy = DV.d2[0];
		B  = V.d0[1];  Bx = V.d1[1];  By = V.d2[1];  DB = DV.d0[1];  DBx = DV.d1[1];  DBy = DV.d2[1];
		om = V.d0[2]; omx = V.d1[2]; omy = V.d2[2]; Dom = DV.d0[2]; Domx = DV.d1[2]; Domy = DV.d2[2];
		mu = V.d0[3]; mux = V.d1[3]; muy = V.d2[3]; Dmu = DV.d0[3]; Dmux = DV.d1[3]; Dmuy = DV.d2[3];

		aux0  = -0.5*x*exp(-4.*u)*grad_grad[GradIndex(2,2)]/B;
		Daux0 = -0.5*exp(-4.*u)/B*(
			grad_grad[GradIndex(2,2)]*(Dx-4.*Du*x-x*DB/B)+x*Dgrad_grad[GradIndex(2,2)]);
		aux1  = -2.*Bx  +  grad_grad[GradIndex(0,1)] +  aux0;
		Daux1 = -2.*DBx + Dgrad_grad[GradIndex(0,1)] + Daux0;

		JDX[Index(par,idom,0,i,j)] = DLaplace[0] + (Daux1-aux1*DB/B)/B;
		JDX[Index(par,idom,1,i,j)] = DLaplace[1] + 2.*DBx;
		JDX[Index(par,idom,2,i,j)] = DLaplace[2] + 4.*Domx
			- (Dgrad_grad[GradIndex(1,2)] - grad_grad[GradIndex(1,2)]*DB/B)/B
			- 4.*Dgrad_grad[GradIndex(0,2)];
		JDX[Index(par,idom,3,i,j)] = DLaplace[3] - 2.*Dmux - 2.*Dux - 2.*(DBx-DB*Bx/B)/B
			+ Dgrad_grad[GradIndex(0,0)] + (Dgrad_grad[GradIndex(0,1)]-grad_grad[GradIndex(0,1)]*DB/B)/B
			+ 0.5*(Daux0-aux0*DB/B)/B;
		
		if (idom == NDOM-1){ // Matter terms if in ring
			double  A1,  A2,  diff,  diff2,  e2V,  e2mu,  hat_vel,  vel2,  h,  p,  eps,  muB;
			double DA1, DA2, Ddiff, Ddiff2, De2V, De2mu, Dhat_vel, Dvel2, Dh, Dp, Deps, DmuB;
			
			A1    = x*exp(-2.*u);                DA1       = exp(-2.*u)*(Dx - 2.*x*Du);
			A2    = B*B*exp(2.*u);               DA2       = 2.*B*exp(2.*u)*(DB+B*Du);
			diff  = par.Omega - om;                Ddiff     = Dpar.Omega - Dom;
			diff2 = diff*diff;                   Ddiff2    = 2.*diff*Ddiff;
			e2V   = A2-A1*diff2;                 De2V      = DA2-(DA1*diff2 + A1*Ddiff2);
			e2mu  = exp(2.*mu);                  De2mu     = 2.*e2mu*Dmu;

			hat_vel  = exp(-2.*u)*diff/B;        Dhat_vel  = exp(-2.*u)*(Ddiff - diff*(2.*Du + DB/B))/B;
			vel2     = x*hat_vel*hat_vel;        Dvel2     = hat_vel*(Dx*hat_vel + 2.*x*Dhat_vel);
			h        = par.V0 - 0.5*log(e2V);    Dh        = Dpar.V0 - 0.5*De2V/e2V;
			eos_quantities(  par, h, &p, &eps, &muB);
			D_eos_quantities(par, h, Dh, &Dp, &Deps, &DmuB);

			aux0        = (eps+p)*(1.+vel2)/(1.-vel2) - 2.*p;
			Daux0       = ((Deps+Dp)*(1.+vel2)+2.*(eps+p)*Dvel2/(1.-vel2))/(1.-vel2) - 2.*Dp;
			aux1        = (eps+p)*hat_vel/(1.-vel2);
			Daux1       = ((Deps+Dp)*hat_vel+(eps+p)*Dhat_vel+aux1*Dvel2/(1.-vel2))/(1.-vel2);

			JDX[Index(par,idom,0,i,j)] -=   4.*Pi*          (De2mu*aux0    + e2mu*Daux0);
			JDX[Index(par,idom,1,i,j)] -=  16.*Pi*          (De2mu*B*p     + e2mu*(DB*p+B*Dp));
			JDX[Index(par,idom,2,i,j)] -= -16.*Pi*exp(2.*u)*(De2mu*B*aux1  + e2mu*(B*Daux1 + aux1*(DB + 2.*B*Du)));
			JDX[Index(par,idom,3,i,j)] -=  -4.*Pi*          (De2mu*(eps+p) + e2mu*(Deps+Dp));
		}
	}
}
// -------------------------------------------------------------------------------
void NonLinStarBoundary(parameters par, int j, derivs_1D w, derivs_1D G, derivs_2D V, double *F)
{
	int idom=NDOM-1, ipot, ns=par.ns[idom], Ns=ns-1, i=Ns;
	double x, u, B, om, A1, A2, diff, diff2, e2V, h;
	coordinate xx, yy;

	coord_st_To_xy(par, idom, i, j, G, &xx, &yy);
	for(ipot=0; ipot<NPOT; ipot++) 
		V.d0[ipot] = w.d0[Index_Star(par, ipot, j)];
	
	x  = xx.d0;
	u  = V.d0[0];      B  = V.d0[1];     om = V.d0[2];

	A1    = x*exp(-2.*u); 
	A2    = B*B*exp(2.*u);
	diff  = par.Omega - om;
	diff2 = diff*diff;
	e2V   = A2-A1*diff2;
	h     = par.V0 - 0.5*log(e2V);      // h    = V0 - V

	F[par.n_2D + j] = h;      // h = 0 to be fulfilled at the star's surface
}
// -------------------------------------------------------------------------------
void LinStarBoundary(parameters par, parameters Dpar, int j, derivs_1D w, derivs_1D G, derivs_2D V,
	derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX)
{// Linear version of 'NonLinStarBoundary'
	int idom=NDOM-1, ipot, i=par.ns[idom]-1;
	double  x,  u,  B,  om,  A1,  A2,  diff,  diff2,  e2V;
	double Dx, Du, DB, Dom, DA1, DA2, Ddiff, Ddiff2, De2V, Dh;
	coordinate xx, yy, Dxx, Dyy;;

	Dcoord_st_To_xy(par, Dpar, idom, i, j, G, DG, &xx, &yy, &Dxx, &Dyy);
	for(ipot=0; ipot<NPOT; ipot++) {
		int indx_S  = Index_Star(par, ipot, j);
		V.d0[ipot]  = w.d0[indx_S];
		DV.d0[ipot] = Dw.d0[indx_S];
	}

	x   = xx.d0;   Dx  = Dxx.d0;
	u   = V.d0[0]; B   = V.d0[1]; om = V.d0[2]; Du = DV.d0[0]; DB = DV.d0[1]; Dom = DV.d0[2];

	A1    = x*exp(-2.*u);                DA1    = exp(-2.*u)*(Dx - 2.*x*Du);
	A2    = B*B*exp(2.*u);               DA2    = 2.*B*exp(2.*u)*(DB+B*Du);
	diff  = par.Omega - om;                Ddiff  = Dpar.Omega - Dom;
	diff2 = diff*diff;                   Ddiff2 = 2.*diff*Ddiff;
	e2V   = A2-A1*diff2;                 De2V   = DA2-(DA1*diff2 + A1*Ddiff2);

	Dh    = Dpar.V0 - 0.5*De2V/e2V;

	JDX[par.n_2D + j] = Dh;      // h = 0 to be fulfilled at the star's surface
}
// -------------------------------------------------------------------------------
void NonLinInfinityEqns(parameters par, int idom, int i, int j, derivs_2D v, double *F)
{// Boundary conditions at infinity
	int indx;	
		
	indx = Index(par,idom,0,i,j); // Potential u, ipot=0
	F[indx] = v.d0[indx];
	
	indx = Index(par,idom,1,i,j); // Potential B, ipot=1
	F[indx] = v.d0[indx] - 1.;
	
	indx = Index(par,idom,2,i,j); // Potential om, ipot=2
	F[indx] = v.d0[indx];

	indx = Index(par,idom,3,i,j); // Potential mu, ipot=3
	F[indx] = v.d0[indx];
}	
// -------------------------------------------------------------------------------
void LinInfinityEqns(parameters par, int idom, int i, int j, derivs_2D Dv, double *JDX)
{// Linear version of 'NonLinInfinityEqns'
	int indx;	
		
	indx = Index(par,idom,0,i,j); // Potential u, ipot=0
	JDX[indx] = Dv.d0[indx];
	
	indx = Index(par,idom,1,i,j); // Potential B, ipot=1
	JDX[indx] = Dv.d0[indx];

	indx = Index(par,idom,2,i,j); // Potential om, ipot=2
	JDX[indx] = Dv.d0[indx];

	indx = Index(par,idom,3,i,j); // Potential mu, ipot=3
	JDX[indx] = Dv.d0[indx];

}	
// -------------------------------------------------------------------------------
void NonLinTransitEqns(parameters par, int idom, int idom1, int idom2, int is, int i, int j, 
	derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F)
{// Transition equations:
 // 1) Calculating the (x,y)-coordinates for given (s,t) (and derivatives)
 // 2) dx,dy: Derivatives with respect to the boundary direction:
 //       Boundary is described by s=1, t varies along boundary
 // 3) Calculating U and the normal derivative U_n (for t=1: U_n*y_t*rp, yp=rp^2)
 //       in (x,y)-coordinates for ipot=0..NPOT-1
 // 4) Both function value and normal derivative is multiplied by (is)
 //    and added to the values
 //       F[Index(par,idom1,ipot,i1,j1)] and F[Index(par,idom2,ipot,i2,j2)] respectively,
 //    and the values (i1,j1) are determined from (i,j) with respect to the location on the boundary
	int nt=par.nt[idom], Nt=nt-1, ipot;
	double normal, xs, xt, ys, yt, yst;
	coordinate xx, yy;

	coord_st_To_xy(par, idom, i, j, G, &xx, &yy);
	xs = xx.d1; xt = xx.d2; ys = yy.d1; yt = yy.d2;  yst = yy.d12;

	v_To_V(par, idom, i, j, v, w, V);        // Calculation of (V, V_s, V_t, V_ss, V_st, V_tt)
	if (j < Nt) pot_st_To_xy(par, xx, yy, V);// Calculation of (V, V_x, V_y) for t<1

	for(ipot=0; ipot<NPOT; ipot++) {
		int i1=par.ns[idom1]-1, j1=j, i2=par.ns[idom2]-1, j2=j,
			indx1 = Index(par,idom1,ipot,i1,j1),
			indx2 = Index(par,idom2,ipot,i2,j2);

		F[indx1] += V.d0[ipot]*is;
		
		if (j < Nt)
			normal = xt*V.d2[ipot] - yt*V.d1[ipot];
		else{// For t=1, choose continuity of (yt*U_n)
			if (idom == 0)
				normal = xt*V.d2[ipot]-(xt*xt+yt*yt)/xs*V.d1[ipot];
			else
				normal = (-yt*yt*V.d2[ipot] + (xt*xt+yt*yt)*V.d12[ipot])/xt;
		}
		F[indx2]  += normal*is;
	}
}
// -------------------------------------------------------------------------------
void LinTransitEqns(parameters par, parameters Dpar, int idom, int idom1, int idom2, int is, int i, int j, 
	derivs_2D  v, derivs_1D  w, derivs_1D  G, derivs_2D  V,
	derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double* JDX)
{// Linear version of 'NonLinTransitEqns'

	int nt=par.nt[idom], Nt=nt-1, ipot;
	double xs, xt, ys, yt, yst, Dnormal, Dxs, Dxt, Dys, Dyt, Dyst;
	coordinate xx, yy, Dxx, Dyy;

	Dcoord_st_To_xy(par, Dpar, idom, i, j, G, DG, &xx, &yy, &Dxx, &Dyy);
	xs  =  xx.d1;  xt =  xx.d2;  ys =  yy.d1;  yt =  yy.d2;  yst =  yy.d12;
	Dxs = Dxx.d1; Dxt = Dxx.d2; Dys = Dyy.d1; Dyt = Dyy.d2; Dyst = Dyy.d12;
	
	v_To_V(par, idom, i, j,  v,  w,  V);
	v_To_V(par, idom, i, j, Dv, Dw, DV);
	if (j < Nt) D_pot_st_To_xy(par, xx, yy, V, Dxx, Dyy, DV);

	for(ipot=0; ipot<NPOT; ipot++) {
		int i1=par.ns[idom1]-1, j1=j, i2=par.ns[idom2]-1, j2=j,
			indx1 = Index(par,idom1,ipot,i1,j1),
			indx2 = Index(par,idom2,ipot,i2,j2);

		JDX[indx1] += DV.d0[ipot]*is;
		
		if (j < Nt){
// 			normal  = xt*V.d2[ipot] - yt*V.d1[ipot];
			Dnormal = (Dxt*V.d2[ipot] + xt*DV.d2[ipot]) - (Dyt*V.d1[ipot] + yt*DV.d1[ipot]);
		}
		else{// For t=1, choose continuity of (y_t*U_n)
			double aux1 = xt*xt+yt*yt, Daux1 = 2.*(xt*Dxt+yt*Dyt);
			if (idom == 0){
				double aux2 = V.d1[ipot]/xs, Daux2 = (DV.d1[ipot]-aux2*Dxs)/xs;
// 				normal  = xt*V.d2[ipot]-(xt*xt+yt*yt)/xs*V.d1[ipot];
				Dnormal = Dxt*V.d2[ipot] + xt*DV.d2[ipot] - (Daux1*aux2 + aux1*Daux2);
			}
			else{
				double 
					aux2  = -yt*yt*V.d2[ipot] + aux1*V.d12[ipot],
					Daux2 = -yt*(2.*Dyt*V.d2[ipot] + yt*DV.d2[ipot]) + Daux1*V.d12[ipot] + aux1*DV.d12[ipot];
// 				normal  = (-yt*yt*V.d2[ipot] + (xt*xt+yt*yt)*V.d12[ipot])/xt;
				Dnormal = (Daux2 - aux2*Dxt/xt)/xt;
			}
		}
		JDX[indx2]  += Dnormal*is;
	}
}
// -------------------------------------------------------------------------------
void NonLinBoundEqns(parameters par, int idom, int i, int j, 
						derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F)
{// The Nonlinear Boundary Equations, see pages 13, 14
	int Ns=par.ns[idom]-1;

	switch(idom){
		case 0:
			if (i==0)
				NonLinInfinityEqns(par, idom, i, j, v, F);
			else{ 
				if(i==Ns) // Transition between domains 0 and 1, values to be subtracted, boundary s=1
					NonLinTransitEqns(par, idom, 1, 0, -1, i, j, v, w, G, V, F); 
				else
					NonLinFieldEqns(par, idom, i, j, v, w, G, V, F);
			}
			break;
		case 1:
			if (i==Ns)   // Transition between domains 0 and 1, values to be added, boundary s=1
				NonLinTransitEqns(par, idom, 1, 0, +1, i, j, v, w, G, V, F);
			else
				NonLinFieldEqns(par, idom, i, j, v, w, G, V, F);
			break;
	}
}
// -------------------------------------------------------------------------------
void LinBoundEqns(parameters par, parameters Dpar, int idom, int i, int j, 
					derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V,
					derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX)
{// Linear Version of 'NonLinBoundEqns'
	int Ns=par.ns[idom]-1;

	switch(idom){
		case 0:
			if (i==0)
				LinInfinityEqns(par, idom, i, j, Dv, JDX);
			else{
				if(i==Ns)// Transition between domains 0 and 1, values to be subtracted, boundary s=1
					LinTransitEqns(par, Dpar, idom, 1, 0, -1, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX); 
				else
					LinFieldEqns(par, Dpar, idom, i, j, v, w, G, V,  Dv, Dw, DG, DV, JDX);
			}
			break;
		case 1:
			if (i==Ns)   // Transition between domains 0 and 1, values to be added, boundary s=1
				LinTransitEqns(par, Dpar, idom, 1, 0, +1, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX);
			else
				LinFieldEqns(par, Dpar, idom, i, j, v, w, G, V,  Dv, Dw, DG, DV, JDX);
			break;
	}
}
// -------------------------------------------------------------------------------
