#include "spheroidal.h"

#define KNEUTRON 0.0111803398875   // =1/(8*5^1.5)
#define P par.eos3_p


// ----------------------------------------------------------
//   Equation of state: density and pressure as they depend on h
//   h(p) = Integrate[1/(eps(pp)+pp), {pp, 0, p}]
// ----------------------------------------------------------
void scan_eos_parameters(parameters *par, FILE *stream)
{
	int eos;
	double d_aux;
	char string[200];

	fscanf(stream, " %d %s ",  &eos, string);     (*par).eos    = eos;

	switch(eos){
		case 0: // for homogeneous body
			{
				fscanf(stream, " %lf %s ",  &d_aux, string);    (*par).eps0  = d_aux; 
				(*par).h0  = 1.;
				break;
			}
		case 1: // for polytropic body
			{
				fscanf(stream, " %lf %s ",  &d_aux, string);    (*par).N_poly  = d_aux;
				fscanf(stream, " %lf %s ",  &d_aux, string);    (*par).K_poly  = d_aux;
				(*par).h0  = 1.;
				break;
			}
		case 2: // for ideal, relativistic neutron gas,
		        // normalization: for p->0 one finds the polytrope with n=3/2
			{
				(*par).h0  = 1.;
				break;
			}
		case 3: // customized equation of state
			{
				int k, Np;
		
				fill0_dvector((*par).eos3_p, 0, NpMAX-1);

				fscanf(stream, " %d  %lf %s ",  &Np, &d_aux, string); 
				(*par).eos3_Np  = Np;     (*par).h0  = d_aux;

				for(k=1; k <= Np; k++){
					fscanf(stream, " %lf %s ",  &d_aux, string);    (*par).eos3_p[k] = d_aux;
// 					printf("k=%d \t p[k]=%e\n", k ,(*par).eos3_p[k]);
				}
				(*par).eos3_muB0  = (*par).eos3_p[1]/(*par).h0;
				break;
			}
	}
}
// ----------------------------------------------------------
void print_eos_parameters(parameters par, FILE *stream)
{
	int eos=par.eos;

	switch(eos){
		case 0: // for homogeneous body
			{
				fprintf(stream, "%d      \t eos_for_homogeneous_stars \n", eos);
				fprintf(stream, "%20.15e \t eps0 \n",  par.eps0);
				break;
			}
		case 1: // for polytropic body
			{
				fprintf(stream, "%d      \t eos_for_polytropic_stars \n", eos);
				fprintf(stream, "%20.15e \t N_poly \n",  par.N_poly);
				fprintf(stream, "%20.15e \t K_poly \n",  par.K_poly);
				break;
			}
		case 2: // for ideal, relativistic neutron gas,
		        // normalization: for p->0 one finds the polytrope with n=3/2
			{
				fprintf(stream, "%d      \t eos_for_ideal_neutron_gas_stars \n", eos);
				break;
			}
		case 3: // customized equation of state
			{
				int k, Np = par.eos3_Np;

				fprintf(stream, "%d  Customized_equation_of_state\n\n", eos);
				fprintf(stream, "%d    %20.15e     Np_and_h0\n\n",  Np, par.h0);
				for(k=1; k <= Np; k++)
					fprintf(stream, "      %20.15e \t eos_p[%d] \n",  P[k], k);

				break;
			}
	}
}
// ----------------------------------------------------------
void display_eos_parameters(parameters par)
{
	int eos=par.eos;

	switch(eos){
		case 0: // for homogeneous body
			{
				printf(" EOS 0: homogeneous stars, eps0= %20.15e \n\n",  par.eps0);
				break;
			}
		case 1: // for polytropic body
			{
				printf(" EOS 1: polytropic stars, N_poly = %e, K_poly = %e\n\n",  par.N_poly, par.K_poly);
				break;
			}
		case 2: // for ideal, relativistic neutron gas,
		        // normalization: for p->0 one finds the polytrope with n=3/2
			{
				printf(" EOS 2: ideal Neutron gas stars \n\n");
				break;
			}
		case 3: // customized equation of state
			{
				int k, Np = par.eos3_Np;

				printf(" EOS 3: Customized equation of state:\n\n");
				printf("   Np = %d \t h0 = %20.15e\n\n",  Np, par.h0);
				for(k=1; k <= Np; k++)
					printf("   eos3_p[%2d] = %20.15e \n",  k, P[k]);
                printf("\n");

				break;
			}
	}
}
//   -----------------------------------------------
double eos3_pressure(parameters par, double H)
{ // p = p(H)
	double p;
	int Np = par.eos3_Np, k;

	p = P[Np];

	for(k=Np-1; k>=1; k--)
		p = p*H + P[k];

	p = p*H;

	return p;
}
//   -----------------------------------------------
double eos3_dpressure(parameters par, double H)
{ // dp/dH
	double dp;
	int   Np = par.eos3_Np, k;

	dp = P[Np]*Np;

	for(k=Np-1; k>=1; k--)
		dp = dp*H + P[k]*k;

	return dp;
}
//   -----------------------------------------------
double eos3_d2pressure(parameters par, double H)
{ // d^2 p / (d H)^2
	double d2p;
	int   Np = par.eos3_Np, k;

	d2p = P[Np]*Np*(Np-1);

	for(k=Np-1; k>=2; k--)
		d2p = d2p*H + P[k]*k*(k-1);

	return d2p;
}
//   -----------------------------------------------
void eos_quantities(parameters par, double h, double *p, double *eps, double *muB)
{
	int eos = par.eos;
	h = fabs(h);

	switch(eos){
		case 0: // eos for homogeneous body
			{
				*p    = par.eps0*(exp(h)-1.);
				*eps  = par.eps0;
				*muB = par.eps0;
				break;
			}
		case 1: //polytropic eos
			{
				double 
					N_poly = par.N_poly, K_poly = par.K_poly,
					H      = exp(h)-1.,    aux1 = H/(K_poly*(N_poly+1.)), aux2 = pow(aux1, N_poly);
				*p    = K_poly*aux1*aux2;
				*eps  = aux2*(1. + N_poly*H/(N_poly+1.));
				*muB = aux2;
				break;
			}
		case 2: // eos for ideal, relativistic neutron gas,
		        // normalization: for p->0 one finds the polytrope with n=3/2
			{
				double eh = exp(h), e2h = eh*eh,
					xFermi = sqrt(e2h-1.),
					h_ln   = log(xFermi+eh);

				*eps = KNEUTRON*((6.*pow(xFermi,3) + 3.*xFermi)*eh - 3.*h_ln);
				*p   = KNEUTRON*((2.*pow(xFermi,3) - 3.*xFermi)*eh + 3.*h_ln);
				*muB = 8.*KNEUTRON*pow(xFermi,3);
				break;
			}
		case 3: // customized equation of state
			{	
				double H = exp(h)-1., dp = eos3_dpressure(par, H);

				*p    = eos3_pressure(par, H); // p = p(H) = P[1]*H + .. + P[Np]*H^(Np)
				*eps  = (1.+H)*dp - (*p); // dh = dH/(1+H) = dp/(eps+p) => eps(H) = (1+H)*dp/dH - p(H)
				*muB  = dp/par.h0; // d muB / muB = d eps / (eps+p) 
				break;                          // => muB = muB0/eps0 * (eps+p)/(1+H) = muB0/eps0 *dp/dH; eps0 = P[1]
			}
	}
}
// ----------------------------------------------------------
void D_eos_quantities(parameters par, double h, double Dh, double *Dp, double *Deps, double *DmuB)
{
	int eos = par.eos;
	h = fabs(h);

	switch(eos){
		case 0: // eos for homogeneous body
			{
				*Dp    = par.eps0*exp(h)*Dh;
				*Deps  = 0.;
				*DmuB  = 0.;
				break;
			}
		case 1: //polytropic eos
			{
				double 
					N_poly = par.N_poly, K_poly = par.K_poly,
					H      = exp(h)-1.,     aux1 = H/(K_poly*(N_poly+1.)),   aux2 = pow(aux1, N_poly),
					DH     = exp(h)*Dh,    Daux1 = DH/(K_poly*(N_poly+1.)), 
					Daux2  = N_poly*pow(aux1, N_poly-1.)*Daux1;
				*Dp    = K_poly*(Daux1*aux2 + aux1*Daux2);
				*Deps  = Daux2*(1. + N_poly*H/(N_poly+1.)) + aux2*N_poly*DH/(N_poly+1.);
				*DmuB = Daux2;
				break;
			}
		case 2: // eos for ideal, relativistic neutron gas,
		        // normalization: for p->0 one finds the polytrope with n=3/2
			{
				double 
					eh      = exp(h),              Deh      = eh*Dh, 
					xFermi  = sqrt(eh*eh-1.),      DxFermi  = eh*Deh/xFermi,
					xFermi3 = pow(xFermi,3),       DxFermi3 = 3.*xFermi*xFermi*DxFermi,
					                               Dh_ln    = (DxFermi+Deh)/(xFermi+eh);

				*Deps = KNEUTRON*((6.*DxFermi3 + 3.*DxFermi)*eh + (6.*xFermi3 + 3.*xFermi)*Deh - 3.*Dh_ln);
				*Dp   = KNEUTRON*((2.*DxFermi3 - 3.*DxFermi)*eh + (2.*xFermi3 - 3.*xFermi)*Deh + 3.*Dh_ln);
				*DmuB = 8.*KNEUTRON*DxFermi3;
				break;
			}
		case 3: // customized equation of state
			{	
				double H = exp(h)-1., DH = (1.+H)*Dh, d2p = eos3_d2pressure(par, H);

				*Dp    = eos3_dpressure(par, H)*DH;
				*Deps  = (1.+H)*d2p*DH;
				*DmuB  = d2p*DH/par.h0;
				break;
			}
	}
}
// ----------------------------------------------------------
#undef KNEUTRON
