#include "spheroidal.h"
#define STENCILSIZE 32

//-----------------------------------------------------------------------------
void resid(parameters par, double *DX, double *rhs, JFD_Components JFD, double *res)
{// Calculates res = rhs - JFD*DX
	int j, k, row, col, ntotal=par.ntotal, num_v=par.num_v, nR=ntotal-num_v;
	
	for(j=0; j<num_v; j++){
		res[j] = 0.;
		for(k=0; k < JFD.ncols_J_UL[j]; k++){
			col     = JFD.cols_J_UL[j][k];
			res[j] -= JFD.J_UL[j][k]*DX[col];
		}
		for(k=0; k<nR; k++){
			col     = num_v+k;
			res[j] -= JFD.J_UR[j][k]*DX[col];
		}
		res[j] += rhs[j];
	}
	for(j=0; j<nR; j++){
		row      = num_v+j;
		res[row] = 0.;
		for(k=0; k<num_v; k++)
			res[row] -= JFD.J_LL[j][k]*DX[k];
		for(k=0; k<nR; k++){
			col       = num_v+k;
			res[row] -= JFD.J_LR[j][k]*DX[col];
		}
		res[row] += rhs[row];
	}
}
// -----------------------------------------------------------------------------------
void Solve_JFD_UL(parameters par, double *F, double *X, JFD_Components JFD)
{// Solves the submatrix-problem J_UL*X = F
	int j, num_v=par.num_v, m1=JFD.m1, m2=JFD.m2;
	double *y;

	y = dvector(1, num_v);

	// Calculating the vector y of eqn. 2.7.20 with A=K_UL, B=F
	for(j=1; j<=num_v; j++) y[j] = F[j-1];
	banbks(JFD.K_UL, num_v, m1, m2, JFD.Kl_UL, JFD.iK_UL, y);
	for(j=1; j<=num_v; j++)
		X[j-1] = y[j];
	free_dvector(y, 1, num_v);
}
// -----------------------------------------------------------------------------------
void Solve_JFD(parameters par, double *F, double *DX, JFD_Components JFD)
{// Solves the preconditioning step JFD*DX = F, see pages 17, 18

	int j, k, ntotal=par.ntotal, num_v=par.num_v, nR=ntotal-num_v;
	double *Y_U, *Y_L;

	Y_U = dvector(0, num_v-1);
	Y_L = dvector(0, nR-1);

	// Calculate Y_U = (J_UL)^(-1)*F^U
	Solve_JFD_UL(par, F, Y_U, JFD);

	// Calculate Y_L = F^L - J_LL*Y_U
	for(j=0; j<nR; j++){
		Y_L[j] = F[num_v+j];
		for(k=0; k<num_v; k++) Y_L[j] -= JFD.J_LL[j][k]*Y_U[k];
	}
	// Calculate DX^L = K_LR*Y_L
	lubksb(JFD.K_LR, nR, JFD.iK_LR, Y_L);
	for(j=0; j<nR; j++) DX[num_v+j] = Y_L[j]; 

	// Calculate DX^U = Y_U - K_UR*DX^L
	for(j=0; j<num_v; j++){
		DX[j] = Y_U[j];
		for(k=0; k<nR; k++) DX[j] -= JFD.K_UR[j][k]*Y_L[k];
//		for(k=0; k<nR; k++) DX[j] -= JFD.K_UR[j][k]*DX[num_v+k];
	}

	free_dvector(Y_U, 0, num_v-1);
	free_dvector(Y_L, 0, nR-1);
}
// -----------------------------------------------------------------------------------
int PreCond(parameters par, double *F, double *DX, double *norm, JFD_Components JFD)
{// Performs and improves the preconditioning step up to <PreCond_itmax> times
 // using the methods of iterative improvements, 'Numerical Recipes in C', pages 55-58
	int iter=0, n, ntotal=par.ntotal;
	double norm0 = norm2(F, ntotal), *res, *XX;

	res=dvector(0,ntotal-1);
	XX =dvector(0,ntotal-1);

	fill0_dvector(DX, 0, ntotal);
	copy_dvector(res, F, 0, ntotal);
	do{
		iter += 1;
		Solve_JFD(par, res, XX, JFD);
		for (n=0; n<ntotal; n++) DX[n] += XX[n];
		resid(par, DX, F, JFD, res);
		*norm = norm2(res, ntotal);
	}
	while(iter < par.PreCond_itmax && *norm > par.PreCond_Improve*norm0);

	free_dvector(res, 0, ntotal-1);
	free_dvector(XX,  0, ntotal-1);

	return iter;
}
// -----------------------------------------------------------------------------------
int bicgstab_Iterations(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters parX, 
		double *F, double *DX, double *normres, JFD_Components JFD)
{	// Iterations for the Biconjugate Stabilized Gradient Method
	int ntotal=par.ntotal, iter, n, N_Pre;
	double norm, alpha = 0., beta = 0., rho = 0., rho1 = 1., rhotol = 1e-50, omega = 0., omegatol = 1e-50,
		*p, *ph, *q, *r, *rt, *s, *sh, *t;

	p  = dvector(0, ntotal-1);		ph = dvector(0, ntotal-1);
	q  = dvector(0, ntotal-1);		r  = dvector(0, ntotal-1);
	s  = dvector(0, ntotal-1);		sh = dvector(0, ntotal-1);
	rt = dvector(0, ntotal-1);		t  = dvector(0, ntotal-1);

	// 	compute initial residual rt = r = p = F - J*DX
	J_times_DX(par, X, DX, v, w, G, parX, r);
	for (n = 0; n < ntotal; n++) rt[n] = r[n] = p[n] = F[n] - r[n];
	
	*normres = norm2(r, ntotal);
	if (*normres <= par.bicgstab_tol) return 0;

	for (iter = 0; iter < par.bicgstab_itmax; iter++) {
		rho = scalarproduct(rt, r, ntotal);
		if (fabs(rho) < rhotol) break;
		if (iter > 0){ // compute direction vector p
			beta = (rho/rho1)*(alpha/omega);
			for (n = 0; n < ntotal; n++) 
				p[n] = r[n] + beta * (p[n] - omega * q[n]);
		}
		if (par.bicgstab_verb == 1) printf(" (%10.3e,", norm1(p, ntotal));
		// compute direction adjusting vector ph and scalar alpha :
		N_Pre = PreCond(par, p, ph, &norm, JFD); 
		if (par.bicgstab_verb == 1) printf("%10.3e) N_Pre=%5d | ", norm, N_Pre);
		J_times_DX(par, X, ph, v, w, G, parX, q);	// q = J*ph
		alpha = rho/scalarproduct(rt, q, ntotal);
		for (n = 0; n < ntotal; n++) s[n] = r[n] - alpha * q[n];
		// early check of tolerance:
		*normres = norm2(s, ntotal);
		if (*normres <= par.bicgstab_tol) {
			for (n = 0; n < ntotal; n++)
				DX[n] += alpha * ph[n];
			if (par.bicgstab_verb == 1) printf("%5d  %10.3e\n", iter+1, *normres);
			break;
		}
		// compute stabilizer vector sh and scalar omega:
		if (par.bicgstab_verb == 1) printf("(%10.3e,", norm1(s, ntotal));
		N_Pre = PreCond(par, s, sh, &norm, JFD);
		if (par.bicgstab_verb == 1) printf("%10.3e) N_Pre=%5d | ", norm, N_Pre);
		J_times_DX(par, X, sh, v, w, G, parX, t);	// t = J*sh
		omega = scalarproduct(t, s, ntotal) / scalarproduct(t, t, ntotal);
		// compute new solution approximation:
		for (n = 0; n < ntotal; n++) {
			DX[n] += alpha * ph[n] + omega * sh[n];
			r[n]   = s[n] - omega * t[n];
		}
		// are we done? 
		*normres = norm2(r, ntotal);
		if (par.bicgstab_verb == 1) printf("%5d  %10.3e\n", iter+1, *normres);
		if (*normres <= par.bicgstab_tol) break;
		rho1 = rho;
		if (fabs(omega) < omegatol) break;
	}
	if (par.bicgstab_verb == 1) printf("\n");

	free_dvector(p,  0, ntotal-1); 		free_dvector(ph, 0, ntotal-1); 
	free_dvector(q,  0, ntotal-1); 		free_dvector(r,  0, ntotal-1);
	free_dvector(s,  0, ntotal-1); 		free_dvector(sh, 0, ntotal-1); 
	free_dvector(rt, 0, ntotal-1); 		free_dvector(t,  0, ntotal-1);

	/* iteration failed */
	if (iter > par.bicgstab_itmax) return -1;
	
	/* breakdown */
	if (fabs(rho) < rhotol) return -10;
	if (fabs(omega) < omegatol) return -11;
	
	/* success! */
	return iter+1;
}
// -----------------------------------------------------------------------------------
int bicgstab(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters parX, double *F, double *DX,
    double *normres)
{	// Biconjugate Stabilized Gradient Method, see pages 1, 2, http://www.netlib.org/templates/
	int ntotal=par.ntotal, num_v=par.num_v, nR=ntotal-num_v, maxcol = NPOT*STENCILSIZE, iter;
	JFD_Components JFD;

	// allocating the components of JFD
	JFD.J_UL         = dmatrix(0, num_v-1, 0, maxcol-1);      fill0_dmatrix(JFD.J_UL,        0, num_v,   0, maxcol);
	JFD.cols_J_UL    = imatrix(0, num_v-1, 0, maxcol-1);      fill0_imatrix(JFD.cols_J_UL,   0, num_v,   0, maxcol);
	JFD.ncols_J_UL   = ivector(0, num_v-1);                   fill0_ivector(JFD.ncols_J_UL,  0, num_v);

	JFD.J_UR         = dmatrix(0, num_v-1, 0, nR-1);          fill0_dmatrix(JFD.J_UR,        0, num_v,   0, nR);
	JFD.K_UR         = dmatrix(0, num_v-1, 0, nR-1);          fill0_dmatrix(JFD.K_UR,        0, num_v,   0, nR);

	JFD.J_LL         = dmatrix(0, nR-1, 0, num_v-1);          fill0_dmatrix(JFD.J_LL,        0, nR,      0, num_v);

	JFD.J_LR         = dmatrix(0, nR-1, 0, nR-1);             fill0_dmatrix(JFD.J_LR,        0, nR,      0, nR);
	JFD.K_LR         = dmatrix(0, nR-1, 0, nR-1);             fill0_dmatrix(JFD.K_LR,        0, nR,      0, nR);
	JFD.iK_LR        = ivector(0, nR-1);                      fill0_ivector(JFD.iK_LR,       0, nR);
	
	Get_JFD_Matrix(par, X, v, w, G, parX, &JFD);

	JFD.K_UL         = dmatrix(1, num_v, 1, JFD.m1+JFD.m2+1); fill0_dmatrix(JFD.K_UL,        1, num_v+1, 1, JFD.m1+JFD.m2+2);
	JFD.Kl_UL        = dmatrix(1, num_v, 1, JFD.m1);          fill0_dmatrix(JFD.Kl_UL,       1, num_v+1, 1, JFD.m1+1);
	JFD.iK_UL        = ivector(1, num_v);                     fill0_ivector(JFD.iK_UL,       1, num_v+1);

	Get_JFD_Components(par, &JFD);
		
	if (par.bicgstab_verb == 1) 
		printf(" bicgstab:  itmax = %d, tol = %e \n", par.bicgstab_itmax, par.bicgstab_tol);

	iter = bicgstab_Iterations(par, X, v, w, G, parX, F, DX, normres, JFD);
	
	// free the components of JFD
	free_dmatrix(JFD.J_UL,        0, num_v-1, 0, maxcol-1);
	free_imatrix(JFD.cols_J_UL,   0, num_v-1, 0, maxcol-1);
	free_ivector(JFD.ncols_J_UL,  0, num_v-1);

	free_dmatrix(JFD.J_UR,        0, num_v-1, 0, nR-1);
	free_dmatrix(JFD.K_UR,        0, num_v-1, 0, nR-1); 

	free_dmatrix(JFD.J_LL,        0, nR-1, 0, num_v-1); 

	free_dmatrix(JFD.J_LR,        0, nR-1, 0, nR-1);
	free_dmatrix(JFD.K_LR,        0, nR-1, 0, nR-1); 
	free_ivector(JFD.iK_LR,       0, nR-1);

	free_dmatrix(JFD.K_UL,        1, num_v, 1, JFD.m1+JFD.m2+1);
	free_dmatrix(JFD.Kl_UL,       1, num_v, 1, JFD.m1);
	free_ivector(JFD.iK_UL,       1, num_v);

	return iter;
}
// -------------------------------------------------------------------
int newton(parameters par, double *X)
{	// Newton Raphson Method, see pages 1, 2
	parameters parX;
	int ntotal=par.ntotal, n_2D=par.n_2D, ns, Ns, nt, Nt, iter=0, j, bicgstab_iter;
	double *F, *DX, norm, normres;
	derivs_1D G, w;	
	derivs_2D v;	
	
	ns = par.ns[NDOM-1]; Ns = ns - 1;
	nt = par.nt[NDOM-1]; Nt = nt - 1;

	F     = dvector(0, ntotal-1);
	DX    = dvector(0, ntotal-1);
	
	allocate_derivs_1D(&G, nt);
	allocate_derivs_1D(&w, nt*NPOT);
	allocate_derivs_2D(&v, n_2D);

	F_of_X(par, X, v, w, G, &parX, F);
	norm = norm2(F, ntotal);

	if (par.Newton_verb == 1){
		printf(" Newton Raphson Method: Initial Residual: \t |F| = %e \n", norm);
		printf(" ------------------------------------------------------------------\n");
	}
 	while(iter < par.Newton_itmin || (norm > par.Newton_tol  && iter < par.Newton_itmax)){
		iter += 1;
		fill0_dvector(DX, 0, ntotal);

		par.bicgstab_tol = par.bicgstab_decr*norm;
		bicgstab_iter    = bicgstab(par, X, v, w, G, parX, F, DX, &normres);

		for (j = 0; j < ntotal; j++) X[j] -= DX[j];

		F_of_X(par, X, v, w, G, &parX, F);
		norm = norm2(F, ntotal);
		if (par.Newton_verb == 1) 
// 			printf(" Newton: iter = %3d \t Number of bicgstab-steps = %3d \t |F| = %e \n",
			if (isinf(norm) || isnan(norm) || norm > 1.0e+10) {
				printf("\n No Convergence of the Newton Raphson Method. Now exiting to system.\n\n");
				exit(1);
			}
			printf(" Newton: iter = %3d \t |F| = %e \n", iter, norm);
	}
	if(norm > par.Newton_tol)
		printf(
		" Newton Raphson Method failed to converge to prescribed tolerance. Now move on to the next sequence element.\n"
		);
	

	free_dvector(F,  0, ntotal-1);
	free_dvector(DX, 0, ntotal-1);
	
	free_derivs_1D(&G, nt);
	free_derivs_1D(&w, nt*NPOT);
	free_derivs_2D(&v, n_2D);

	return iter;
}
//-----------------------------------------------------------------------------

